/*
 * ocache_disk.c --- disk handling procedures for object caching at KOTETU.
 *      by k-chinen@is.aist-nara.ac.jp, 2000, 2001.
 *
 *  this routines is splitted from xsk.c of kotetu-1.2.
 *
 * $Id$
 */
/*
 * xsk.c --- sk.c + skcache.c
 *      by k-chinen@is.aist-nara.ac.jp, 2000, 2001.
 *
 * $Id$
 */
/*
 * sk.c --- skip list procedures
 *              for KOTETU
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 */
/*
 * skcache.c --- object cache procedures using skip list and its bucket
 *              for KOTETU
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 */

/*

 terminology:
  entry   - object entry. store properties like PCB of TCP/IP
            or super block of filesystem.
  frag    - fragment of object cache. object cache is devided little block
            and stored its list.
            you can access entity of object by drawing its list.
  bucket  - list of entry.  it was controled as skip list.
 
*/
/*
 NOTE:

 hash:
    this program use meny hash to assign object to structures in several level
    and identify among objects.

 on the memory:
    cache consists set of buckets and frag lists in entry.

        bucket[0]
            entry[0]    -> frag[0], frag[1], frag[2] ...
            entry[1]
            entry[2]
            ...

        bucket[1]
            entry[0]
            entry[1]
            entry[2]
            ...

 on the disk:
    cache consists dirctories according to hash.

        aa/bb/cc-dd
        
        aa - first level dir. with first hash (so it is bucket)
        bb - second level dir. with second hash
        cc-dd - file with third and fourth hash.

 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/uio.h>

#include <dirent.h>

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include <sys/mman.h>

#include <assert.h>
#include <errno.h>


#include "range.h"
#include "msg.h"
#include "trace.h"
#include "misc.h"
#include "mem.h"

#include "ocache.h"
#include "sim_stat.h"
#include "sim_hist.h"
#include "mlist.h"

#ifndef DISK_RETRY
#define DISK_RETRY (13)
#endif

#ifndef BODY_FILENAME0
#define BODY_FILENAME0      'B'
#endif
#ifndef BODY_FILENAME 
#define BODY_FILENAME       "BODY"
#endif

#ifndef INDEX_FILENAME0
#define INDEX_FILENAME0     'I'
#endif
#ifndef INDEX_FILENAME  
#define INDEX_FILENAME      "INDEX"
#endif


#define H4_DIR_COUNT_MAX    (128)
#define H4_DIR_SIZE_MAX     (128*K(10))
/*
#define H4_DIR_SIZE_MAX     (K(52))
*/

volatile int dir_count_max      = H4_DIR_COUNT_MAX;
volatile off_t dir_size_max     = H4_DIR_SIZE_MAX;
volatile off_t dir_size_hwater  = H4_DIR_SIZE_MAX;
volatile off_t dir_size_lwater  = H4_DIR_SIZE_MAX/2;

extern int require_blksize;
extern int require_nblk_file;
extern int require_nblk_seek;
extern int require_nblk_obj;

/*
 * arround 256, follwing numbers are known as prime numbers.
 *  241 251 257 263 269 271
 *
 */

extern int radix_h1;
extern int radix_h2;
extern int radix_h3;
extern int radix_h4;

extern skcache_entry_t *NIL;
extern int nbucket;
extern bucket_t *skcache_bucket_base;

extern pthread_mutex_t stat_lock;

volatile int count_too_big_object = 0;

volatile int swapin_count       = 0;
volatile int swapin_success     = 0;
volatile int swapout_count      = 0;
volatile int swapout_success    = 0;

volatile int checkdir_dir_count         = 0;
volatile int checkdir_file_read_count   = 0;
volatile int checkdir_file_handle_count = 0;
volatile int checkdir_file_left_count   = 0;
volatile int checkdir_file_delete_count = 0;
volatile int checkdir_file_erase_count  = 0;

volatile int check_dir_ignore_hash_count = 0;

sim_stat_rec    swapin_time;
static long     swapin_time_count[K(1)];
sim_stat_rec    swapout_time;
static long     swapout_time_count[K(1)];

sim_hist_rec    swapin_hist;
static long     swapin_hist_area[K(1)];
sim_hist_rec    swapout_hist;
static long     swapout_hist_area[K(1)];

#ifndef FS_HIST_LEN
#define FS_HIST_LEN (1024)
#endif

volatile int fs_swapin_count[RADIX_H1_MAX];
volatile int fs_swapout_count[RADIX_H1_MAX];

sim_hist_rec    fs_swapout_hist[RADIX_H1_MAX];
static long     fs_swapout_hist_area[RADIX_H1_MAX][FS_HIST_LEN];
sim_hist_rec    fs_swapin_hist[RADIX_H1_MAX];
static long     fs_swapin_hist_area[RADIX_H1_MAX][FS_HIST_LEN];

sim_stat_rec    checkdir_time;
static long     checkdir_time_count[K(2)];
sim_stat_rec    checkdir_time_short;
static long     checkdir_time_short_stat_count[K(2)];

sim_stat_rec    swapout_fail_reason;
static long     swapout_fail_reason_count[K(1)];
sim_stat_rec    swapin_fail_reason;
static long     swapin_fail_reason_count[K(1)];


extern int require_radix_h2;
extern int require_radix_h3;
extern off_t require_dir_size;
#if 0
extern int require_swap_token_val;
extern int require_swap_token_interval;
#endif
extern int require_keeptime_token_val;
extern int require_disk_sweep_interval;



extern pthread_mutex_t config_lock;

extern int skcache_entry_keep_cur;


volatile long long swap_count = 0LL;
volatile long long swap_size = 0LL;

/* limitter */
#if 0
volatile int swapout_token_init = 1000;
volatile int swapout_token  = 1000;
volatile int swapout_token_interval;
#endif
volatile int disk_sweep_interval;

time_t min_date;

extern pthread_mutex_t dir_mutex;
extern int dir_num;
extern int *dir_lock;
extern u_int32_t *dir_sig;
extern off_t *dir_sz;
extern int *dir_count;
extern int *dir_time;
extern int *dir_lack;
extern int dir_conflict;


u_int32_t *dir_bit = NULL;
int *dir_map = NULL;
int bitmapping_flag = 1;


volatile int fszchk = 1;

extern int nindex_dread;
extern int nindex_bdread;
extern int nindex_bdwrite;


TRACE_REGIST_NIC("objcachedisk");


#if 1
/* nanosleep is useful, because it is MT-safe and supports very short time.
 * however, nsec is too short for network I/O.
 * then, we made msleep.  So, 1/1000 sec scale sleep.
 */
/*
inline
*/
static
int
msleep(int x)
{
    struct timespec w;
    long long tw;

    /* convert msec to nsec */
    tw = x * 1000000;
    w.tv_nsec = tw % 1000000000;
    w.tv_sec =  tw / 1000000000;

    nanosleep(&w, NULL);
    
    return 0;
}
#endif


#ifndef HAVE_DIRENT_T
typedef struct dirent dirent_t;
#endif


#include "ocache_fsmap.c"
#include "ocache_low.c"



/* 
 * XXX 'first match'
 *
 * seek bitmap on memory
 *
 *
 * return:
 *      -1  not found or error
 *      0>  found and its position
 *              for example, the fuction returns 3.
 *              blocks between 3 to 3+want-1 are free.
 *              the caller should use them for new swap.
 */
static
int
_psf_bit_firstmatch(int sig_pos, int want, int npos, int verbose)
{
    int ii, i;
    int v;
    int match;
    int first;
    int try;
    int ret;

    ret = -1;

#if 0
    for(i=npos;i<MIN(npos+Onblk_seek, Onblk_file);i++) {
        v = DBIT_ISSET(sig_pos, i);
        Trace("psf: %3d %4d bit %d \n", xid, i, v);
    }
#endif

    try = MIN(Onblk_seek, Onblk_file); /* XXX */
    match = 0;
    first = -1;
    for(ii=npos;ii<npos+Onblk_seek;ii++)
    { 
        i = ii%Onblk_file;

        if(i==0) {
            match = 0;
            first = -1;
        }

        v = DBIT_ISSET(sig_pos, i);

        if(v==0) {
            if(first<0) {
                first = i;
            }
            match++;
        }
        else {
            match = 0;
            first = -1;
        }

 if(verbose) {
Trace("  :F dir# %d, pos %4d %4d, try %3d: v %d match %2d first %4d\n",
        sig_pos, i, ii, try, v, match, first);
 }

        if(match>=want) {
            break;
        }

        try--;
    }

    if(match>=want) {
        ret = first;
    }
    else {
        ret = -1;
    }
 if(verbose) {
Trace("  : dir# %d, ; return %d\n", sig_pos, ret);
 }

    return ret;
}

#ifndef BMRADIX 
#define BMRADIX (53)
#endif
#define DFC ((e->h1)%BMRADIX==1)

/* print debug information of several file by select 1/53 of all sessions
 * DFC is masking for such selection.
 */

/*
 * ifd_seekfree - seek free block chunk on INDEX file.
 *      The chunk have to no break.  Moreover its length is longer
 *      or equal than the target object.
 *
 * return:
 *  success     0
 *  error       <0
 *
 */
int
ifd_seekfree(int sfd, bucket_t *bk, skcache_entry_t *e, int *rpos)
{
    int ret;
    int c;
    off_t pos, dst;
    int fnum;
    int ipos, npos;     /* position by blocks */
    int nr;
    skcache_entry_t *tmp_entry;
    int match = 0;
    int try;
    off_t vpos;
    time_t diff;
    skcache_entry_t chunk[HARD_NINDEX_DREAD];
    int i;
    int nwid, wid;
    int first;
    int want;
    int xid; /* loop id */
    int sig_pos;

    ret = 0;
    *rpos = (off_t) -1;
    xid = rand()%1000;  /* make temporary loop id */
    
    if(e==NULL) {
    }

    fnum = FDPOOL_POS_E(e);
    sig_pos = DSIG_POS(e->h1, e->h2, e->h3);
    want = NBCEIL(e->totallength, Oblksize);

    npos = ipos = ofile_offset(e);
    pos = (off_t)npos * ESIZE;
    POSCHECKI(pos);
    if(pos<0 || npos<0) {
        Error("psf: 1 Ignore position block# %d %lldB; e h3 %d h4 %d; slot %d",
            npos, (long long)pos, e->h3, e->h4, Onblk_file);
        ret = XSK_BODY_EPOS;
        goto out;
    }

#if 0
    if(npos+want-1<=Onblk_file-1) {
        goto not_found;
    }
#endif

#if 0
    if(npos+want-1>Onblk_file-1) {
        goto not_found;
    }
#endif

#if 0
    Trace("psf: 0 [%03d:%03d:%03d:%04d] len %ld/%ld\n",
        e->h1, e->h2, e->h3, e->h4, (long)e->totallength, (long)Oblksize);
#endif

#if 0
 if(DFC) {
Trace("psf: 2 sfd# %d, [%03d:%03d:%03d:%04d] x %6ld => pos %8d/%d %12lldB (%4ldMB+%7ld)\n",
    sfd,
    e->h1, e->h2, e->h3, e->h4, Oblksize,
    npos, Onblk_file,
    (long long)pos, (long)dM(pos), (long)mM(pos) );
 }
#endif

 if(DFC) {
Trace("psf: 2 dir# %d, sfd# %d, %d:%d:%d:%d from pos %d with %d blocks ; count %d\n",
    sig_pos, sfd, e->h1, e->h2, e->h3, e->h4, npos, want, dir_count[sig_pos]);
 }



#if 1
    /*
     * bitmapping is enable.
     * hook position scanning.
     */
    if(bitmapping_flag) {
        if(dir_count[sig_pos]<0) {
#if 0
            Trace("psf: %3d don't ref. dir_bit\n", xid);
#endif

            /* XXX, currently this disk seeking routine does not support
               multi block seeking */

#if 0
            if(want>1) {
                goto not_found;
            }
#endif
#if 0
            if(want>1) {
                Trace("  oid# %d, want %d\n", e->oid, want);
            }
#endif

        }
        else {
            int tmp_pos;

            tmp_pos = _psf_bit_firstmatch(sig_pos, want, npos, DFC);
#if 0
            Trace("result by first-match %d\n", tmp_pos);
#endif

            if(tmp_pos<0) {
                goto not_found;
            }
            else {
                npos = tmp_pos;
                /* XXX */
                pos = (off_t)npos * ESIZE;
                goto found_by_membits;
            }
        }
    }
    else {
    }
#endif

    POSCHECKI(pos);

    if(pos<(off_t)0L) {
        Error("psf: ignore position block# %d %lldB before seek()", 
            npos, (long long)pos);

        ret = XSK_BODY_EPOS;
        goto out;
    }

#if 0
    if(POSVERIFY(sfd, ifd_pos[sig_pos])) {
        ifd_pos[sig_pos] = -1;
    }
#endif

    c = 0;
try_seek:
    c++;
    dst = Iseek(sig_pos, sfd, pos);
    if(dst<(off_t)0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_seek;
        }
Error("psf: cannot seek %d %lldB on index file fd# %d :%d (%d)",
            npos, (long long)pos, sfd, c, errno);

        ret = XSK_BODY_EPOS;
        goto out;
    }
    vpos = dst;
    POSVERIFY(sfd, vpos);
    POSCHECKI(dst);

    try = Onblk_seek;
    match = 0;
    first = -1;
    do {
#if 0
        if(npos+want-1>=Onblk_file) {
            match = 0;
            break;
        }
#endif

        pos = (off_t)npos * ESIZE;
        nwid = MIN(npos+ MIN(nindex_dread, Onblk_seek),
                    Onblk_file) - npos;
        wid = ESIZE*nwid;

        /* block read */

 if(DFC) {
Trace("psf: 4 sfd# %d, pos %d %lld, addr %p, wid %d %d => %d/%d %lld/%lld\n",
    sfd, npos, (long long)pos, chunk, nwid, wid, npos+nwid, Onblk_file,
    (long long)(pos+wid), (long long)(Onblk_file*ESIZE) );
 }

        c = 0;
try_read:
        c++;

        nr = Iread(sig_pos, sfd, chunk, wid);
        if(nr<0) {
            if(errno==EINTR && c < DISK_RETRY) {
                goto try_read;
            }
            Error("psf: cannot read fd# %d: %d (%d)",
                sfd, c, errno);
            break;
        }
        vpos += nr;
        POSVERIFY(sfd, vpos);
        nwid = nr/ESIZE;

#if 0
        for(i=0;i<nwid && try && match<want; i++) 
#endif
        for(i=0;i<nwid && try ; i++) 
        {
            pos = (off_t)npos * ESIZE;
            tmp_entry = &chunk[i];

            if(npos==0) {
                match = 0;
                first = -1;
            }

            /*
             * ignore magic, empty or very older is counted as target
             */
            if((tmp_entry->magic[0]!=MAGIC_0 || tmp_entry->magic[1]!=MAGIC_1)
                || tmp_entry->oid == 0 || tmp_entry->issue_time < min_date) {

                match++;
                if(first<0) {
                    first = npos;
                }
            }
            else {
                match = 0;
                first = -1;
            }

#ifdef TRACE_INDEX
#endif
 if(DFC) {
        diff = tmp_entry->issue_time - min_date;

#if 0
Trace("  :f dir# %d, fd# %d, try %3d: pos %4d %12lldB : vpos %12lldB : magic(hex) %02x %02x, oid# %d, diff %d; match %d, first %d\n",
        sig_pos, sfd, try, npos, (long long)pos, (long long) vpos,
        (unsigned char)tmp_entry->magic[0], (unsigned char)tmp_entry->magic[1],
        tmp_entry->oid, diff, match, first
        );
#endif

Trace("  :f dir# %d, ? pos %4d %4d, try %3d: %d: %02x %02x, oid# %d, hash %d:%d:%d:%d diff %d\n",
        sig_pos, npos, npos, try,
        match,
        (unsigned char)tmp_entry->magic[0], (unsigned char)tmp_entry->magic[1],
        tmp_entry->oid, 
        tmp_entry->h1, tmp_entry->h2, tmp_entry->h3, tmp_entry->h4,
        diff
        );

 }

#if 1
            if(match>=want) {
                break;
            }

            npos = (npos+1)% Onblk_file;
            try--;

            if(npos==0) {
                break;
            }
#else

            if(npos<Onblk_file-1) {
                npos++;
            }
#endif

        }
        if(match>=want) {
            break;
        }
        
        if(npos==0) {
#ifdef TRACE_INDEX
            Trace("  : rewind sfd# %d,\n", sfd);
#endif
            c = 0;
try_rewind:
            c++;
            dst = Iseek(sig_pos, sfd, 0);
            if(dst<(off_t)0) {
                if(errno==EINTR && c<DISK_RETRY) {
                    goto try_rewind;
                }
    Error("psf: cannot rewind %lld on index file fd# %d :%d (%d)",
        (long long)pos, sfd, c, errno);
            }
            vpos = dst;
            POSVERIFY(sfd, vpos);
            POSCHECKI(dst);
        
            match = 0;
            first = -1;
        }
    } while (match==0 && try>0);


    /* 
     * find by disk seeking
     *  - move file pointer to first block on INDEX file
     *  - set return code and position
     */

#if 0
    if(match>=want && first>=0)
#endif
    if(match && first>=0)
    {
#if 0
    Trace("psf: %3d found in disk; block# %d + %d\n", xid, npos, want);
#endif
        POSCHECKI(ifd_pos[sig_pos]);

        c = 0;
try_back:
        c++;
        npos = first;
        pos = first*ESIZE;
        dst = Iseek(sig_pos, sfd, pos);
        if(dst<(off_t)0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_back;
            }
    Error("psf: cannot back %lld on index file fd# %d :%d (%d)",
        (long long)pos, sfd, c, errno);

            ret = XSK_BODY_EROLL;
            *rpos = npos;
            goto out;
        }
        vpos = dst;
        POSVERIFY(sfd, vpos);
        POSCHECKI(dst);

#if 0
        ret = npos;
#endif
        ret = 0;
        *rpos = npos;

        goto found;
    }
    else {
#if 0
Trace("psf: 1 sfd# %d, not found (want %d), [%d:%d:%d:%d] fnum %d, npos %d\n",
        sfd, want, e->h1, e->h2, e->h3, e->h4, FDPOOL_POS_E(e), ofile_offset(e));
#endif
        goto not_found;
    }

    goto out;


    /* 
     * find by bitmask on memory seeking
     *  - move file pointer to first block on INDEX file
     *  - set return code and position
     */

found_by_membits:
#if 0
    Trace("psf: %3d found in mem; block# %d + %d\n", xid, first, want);
#endif
    pos = (off_t)npos * ESIZE;

    c = 0;
try_pos3:
    c++;
    dst = Iseek(sig_pos, sfd, pos);
    if(dst<(off_t)0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_pos3;
        }
Error("psf: cannot seek %lld on index file fd# %d :%d (%d)",
    (long long)pos, sfd, c, errno);

        ret = XSK_BODY_EPOS;
        *rpos = npos;
        goto out;
    }
    vpos = dst;
    POSVERIFY(sfd, vpos);
    POSCHECKI(dst);

    ret = 0;
    *rpos = npos;

    goto found;

found:
 if(DFC) {
    Trace(
"psf: 5 dir# %d, fd# %d, want %d, hash %d:%d:%d:%d range [%d,%d) ; return %d found [%d..%d)\n",
    sig_pos, sfd, want, e->h1, e->h2, e->h3, e->h4,
    ofile_offset(e), (ofile_offset(e)+Onblk_seek)%Onblk_file,
    ret, *rpos, (*rpos+want)%Onblk_file);
 }
    goto out;

not_found:
    dir_lack[fnum]++;
    *rpos = npos;
    ret = XSK_INDEX_NOTFOUND;

 if(DFC) {
    Trace(
"psf: 6 dir# %d, fd# %d, want %d, hash %d:%d:%d:%d range [%d,%d) ; return %d not-found [%d..%d)\n",
    sig_pos, sfd, want, e->h1, e->h2, e->h3, e->h4,
    ofile_offset(e), (ofile_offset(e)+Onblk_seek)%Onblk_file,
    ret, *rpos, (*rpos+want)%Onblk_file);
 }

    goto out;


out:

#if 0
    /* debug for edge of array */

    if(ipos >= Onblk_file - Onblk_obj - 2) {
        Trace(
"psf: dir# %d, fd# %d, ; scan %d..%d, want %d => out %d..%d (%d) with rcode %d\n",
        sig_pos, sfd,
            ipos, ipos+Onblk_seek, want,
            *rpos, *rpos+want, first, ret);
    }

#endif

#if 0
#endif
#ifdef TRACE_INDEX
    pos = (off_t)*rpos * ESIZE;
Trace(
"psf: 7 [%03d:%03d:%03d:%04d] sfd# %d, try %d, match %d, pos %d %12lld <%12lld>, return %d\n",
    e->h1, e->h2, e->h3, e->h4, sfd, Onblk_seek - try, match,
    *rpos, (long long) pos, (long long) vpos, ret);
#endif

    return ret;
}

int
_pso_match(int sig_pos, int start, int value, int verbose)
{
    int ret;
    int i;
    int xv;

    ret = -1;

#if 1
    if(!bitmapping_flag) {
        Error("_pso_match: Don't call me.  Bitmapping is disabled.");
        return -1;
    }
#endif

#if 0
 if(verbose) {
    Trace(" :pso dir# %d, value %d\n", sig_pos, value);
 }
#endif

    for(i=start;i<Onblk_file;i++) {
        xv = DMAP_V(sig_pos, i);
#if 0
        if(verbose)
            Trace(" omap: dir# %d, %4d: %10d\n", sig_pos, i, xv);
#endif
        if(xv==value) {
#if 0
            if(verbose)
                Trace("   found value %d in dir# %d, %d\n", value, sig_pos, i);
#endif
            ret = i;
            break;
        }
    }

 if(verbose) {
#if 0
    Trace("  :O dir# %d, value %d found in pos %d ; use %d unuse %d %.2f%%\n",
        sig_pos, value, ret, u, n, 100.0*u/(u+n));
#endif
    Trace("  :O dir# %d, value %d found in pos %d\n",
        sig_pos, value, ret);
 }

    return ret;
}


/*
 * return:
 *      found       0
 *      not found   >0
 *      error       <0
 */

int
bfd_seekobj(int *rfd, bucket_t *bk, skcache_entry_t *e,
    int *rpos, skcache_entry_t *re)
{
    int ret;
    int sfd;
    int c;
    off_t pos, dst;
    int fpos;
    int npos;
    int nr;
    skcache_entry_t *tmp_entry;
    int match = 0;
    int try;
    off_t vpos;
    time_t diff;
    skcache_entry_t chunk[HARD_NINDEX_DREAD];
    int wid, nwid;
    int i;
    int use_map=0;

    int sig_pos;
    int cret;
    int chk;

/*
    int skip;
*/


    *rpos = (off_t)-1;
    ret = 0;
    sfd = *rfd;
    fpos = FDPOOL_POS_E(e);
    sig_pos = DSIG_POS(e->h1, e->h2, e->h3);

#if 0
    Trace("bfd_seekobj: [%03d:%03d:%03d:%04d] len %ld/%ld\n",
        e->h1, e->h2, e->h3, e->h4, (long)e->totallength, (long)Oblksize);
#endif

 if(DFC) {
Trace(
"pso: dir# %d, sfd# %d, %d:%d:%d:%d from pos %d ; count %d\n",
    sig_pos, sfd, e->h1, e->h2, e->h3, e->h4, npos, dir_count[sig_pos]);
 }

#if 1
    if(bitmapping_flag) {
        if(dir_count[sig_pos]<0) {
            /* nothing */
            /* following to disk seek */
            npos = ofile_offset(e);
        }
        else
        if(dir_count[sig_pos]==0) {
            ret = XSK_INDEX_NOTFOUND;
            *rpos = -1;
            goto out;
        }
        else {
            use_map = 1;

            npos = _pso_match(sig_pos, 0, OMAP_FUNC(e), DFC);
            if(npos >=0) {
                /* NOTE: found such object.  however it is only compared h4.
                 * you should compare other attributes.
                 * and maybe exist more matched objects
                 */
            }
            else {
#if 0
                ret = XSK_INDEX_NOTFOUND;
                *rpos = -1;
                goto out;
#endif
                goto not_found;
            }
        }
    }
    else {
        npos = ofile_offset(e);
        use_map = 0;
    }
#endif


    try = Onblk_seek;
    match = 0;
    do {
        pos = (off_t)npos * ESIZE;
        nwid = MIN(npos+nindex_dread, Onblk_file) -  npos;
        wid = nwid * ESIZE;
        Iseek(sig_pos, sfd, pos);


        /* block read */
        c = 0;
try_read:
        c++;
 if(DFC) {
Trace("  : dir# %d, %c pos %4d %4d, try %3d: wid %d\n",
            sig_pos, (use_map ? 'm':'-'), npos, npos%Onblk_file, try, nwid);
 }

        nr = Iread(sig_pos, sfd, chunk, wid);
        if(nr<0) {
            if(errno==EINTR && c < DISK_RETRY) {
                goto try_read;
            }
            Error("bfd_seekobj: cannot read fd# %d pos %lld: %d (%d)",
                sfd, (long long)pos, c, errno);
        }
        vpos += nr;
        nwid = nr/ESIZE;

        for(i=0;i<nwid && try;i++) {
            pos = (off_t)npos * ESIZE;
            tmp_entry = &chunk[i];

#ifdef TRACE_INDEX
#endif
 if(DFC) {
            diff = tmp_entry->issue_time - min_date;

Trace("  :o dir# %d, %c pos %4d %4d, try %3d: %d: %02x %02x, oid# %d, hash %d:%d:%d:%d diff %d\n",
            sig_pos, (use_map ? 'm':'-'),
            npos, npos%Onblk_file, try,
            DBIT_ISSET(sig_pos, npos),
            (unsigned char)tmp_entry->magic[0],
            (unsigned char)tmp_entry->magic[1],
            tmp_entry->oid,
            tmp_entry->h1, tmp_entry->h2, tmp_entry->h3, tmp_entry->h4,
            diff
            );
 }
            if(tmp_entry->magic[0]==MAGIC_0 && tmp_entry->magic[1]==MAGIC_1) {
                if(tmp_entry->h3==e->h3 && tmp_entry->h4==e->h4) {
                    cret = strcmp(tmp_entry->name,e->name);
 if(DFC) {
                    if(cret!=0) {
                        Trace("  :mem  '%s'\n  :disk '%s' => %d\n",
                            e->name, tmp_entry->name, cret);
                    }
 }
                    if(cret==0) {
                        *re = *tmp_entry;   /* copy */
                        match = 1;
                        break;
                    }
                }
            }

next_blk:
            if(use_map) {
                npos = _pso_match(sig_pos, npos+1, OMAP_FUNC(e), DFC);
                if(npos<0) {
                    goto not_found;
                }
            }
            else {
                npos = (npos+1)% Onblk_file;
            }

            try--;
#if 0
            if(npos==0) {
                break;
            }
#endif
        }
        if(match || try<=0) {
            break;
        }
    
        if(npos==0) {
#ifdef TRACE_INDEX
            Trace("  : rewind sfd# %d,\n", sfd);
#endif

            c = 0;
try_rewind:
            c++;
            dst = Iseek(sig_pos, sfd, 0);
            if(dst<(off_t)0) {
                if(errno==EINTR && c<DISK_RETRY) {
                    goto try_rewind;
                }
    Error("bfd_seekobj: cannot rewind %lld on index file fd# %d :%d (%d)",
        (long long)pos, sfd, c, errno);
                break;
            }
            vpos = dst;
        }
    } while (match==0 && try>0);

    if(match) {
found:
#if 0
Trace("bfd_seekobj: sfd# %d, found, [%d:%d:%d:%d] npos %d, pos %lld, vpos %ld\n",
    sfd, e->h1, e->h2, e->h3, e->h4,
    npos, (long long)pos, (long long) vpos);
#endif

        c = 0;
try_back:
        c++;
        dst = Iseek(sig_pos, sfd, pos);
        if(dst<(off_t)0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_back;
            }
    Error("bfd_seekobj: cannot back %lld on index file fd# %d :%d (%d)",
        (long long)pos, sfd, c, errno);

            ret = XSK_INDEX_EROLL;
            *rpos = npos;
            goto out;
        }
        vpos = dst;

        ret = 0;
        *rpos = npos;

#if 0
#endif
 if(DFC) {
Trace(
"pso: dir# %d sfd# %d, hash %d:%d:%d:%d range [%d..%d) ; return %d found %d\n",
        FDPOOL_POS_E(e), sfd, e->h1, e->h2, e->h3, e->h4,
        ofile_offset(e), (ofile_offset(e)+Onblk_seek)%Onblk_file, ret, *rpos);
 }
    }
    else {
not_found:
        ret = XSK_INDEX_NOTFOUND;
        *rpos = npos;
#if 0
#endif
 if(DFC) {
Trace(
"pso: dir# %d sfd# %d, hash %d:%d:%d:%d range [%d..%d) ; return %d not-found %d\n",
        FDPOOL_POS_E(e), sfd, e->h1, e->h2, e->h3, e->h4,
        ofile_offset(e), (ofile_offset(e)+Onblk_seek)%Onblk_file, ret, *rpos);
 }
    }

    goto out;


out:

#ifdef TRACE_INDEX
    pos = (off_t)*rpos * ESIZE;
Trace("bfd_seekobj: [%03d:%03d:%03d:%04d] *rfd# %d, try %d, match %d, pos %d %12lld <%12lld>, return %d\n",
    e->h1, e->h2, e->h3, e->h4, *rfd, Onblk_seek - try, match,
    *rpos, (long long) pos, (long long) vpos, ret);
#endif

    return ret;
}




/* 
    return:
        1   fail
        0   success
 */
int
write_attach_fd(int *rifd, int *rbfd, int *rnpos,
    bucket_t *bk, skcache_entry_t *e)
{
    int ret;
    int chk;
    int fpos;
    int npos;
    off_t bipos, bppos; 
    int ifd, bfd;
    int c;

    ret = 0;
    fpos = FDPOOL_POS_E(e);

    *rifd = *rbfd = -7;
    *rnpos = -1;
    ifd = bfd = -3;

    chk = ifd_open(&ifd, bk, e, fszchk);
    if(chk) {
        Trace("write_attach_fd: fail open ifd# %d, chk %d\n", ifd, chk);
        *rifd = -1;
        ret = chk;

        goto out;
    }

    *rifd = ifd;

    chk = ifd_seekfree(ifd, bk, e, &npos);
    if(chk) {
#if 1
        if(chk==57) {
        Trace("write_attach_fd: not found in dir# %d, [%d,%d,%d,%d] <%d>\n",
            fpos, e->h1, e->h2, e->h3, e->h4, chk);
        }
        else {
        Trace("write_attach_fd: some error on dir# %d, [%d,%d,%d,%d] <%d>\n",
            fpos, e->h1, e->h2, e->h3, e->h4, chk);
        }
#endif
        *rnpos = -1;
        ret = 1;

        goto out;
    }
    *rnpos = npos;
    bipos = npos*ESIZE;
    bppos = npos*Oblksize;

    POSCHECKI(bipos);
    POSCHECKP(bppos);

    chk = bfd_open(&bfd, bk, e, fszchk);
    if(chk) {
        Error("write_attach_fd: fail open bfd# %d, chk %d", bfd, chk);
        *rbfd = -1;
        ret = 1;

        goto out;
    }

    *rbfd = bfd;


#ifdef NP
    bppos = (off_t)0L;
#endif

    c = 0;
try_seek:
    c++;
    chk = Pseek(fpos, bfd, bppos);
    if(chk<0) {
Error("write_attach_fd: fail seek at npos %ld %lldB in bfd# %d, chk %d (%d)",
(long)npos, (long long)bppos, bfd, chk, errno);
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_seek;
        }
        *rnpos = -1;
        ret = 1;
    }
    else {
        ret = 0;
    }

out:
#if 0
Trace("write_attach_fd: done *rifd# %d, *rbfd# %d, pos %ld %lldB, ret %d\n",
        *rifd, *rbfd, (long)*rnpos, (long long)bppos, ret);
#endif

    return ret;
}



/* 
    return:
        1   fail
        0   success
 */

int
read_attach_fd(int *rifd, int *rbfd, int *rnpos,
    bucket_t *bk, skcache_entry_t *e, skcache_entry_t *re)
{
    int ret;
    int chk;
    off_t vpos;
    int fpos;
    int npos;
    int ifd, bfd;
    off_t bpos;
    int c;

    ret = 0;
    fpos = FDPOOL_POS_E(e);


    *rifd = *rbfd = -25;
    *rnpos = -1;
    
    ifd = bfd = -57;

    chk = ifd_open(&ifd, bk, e, fszchk);
    if(chk) {
        Error("read_attach_fd: fail open ifd# %d, chk %d", ifd, chk);

        *rifd = -1;
        ret = chk;

        goto out;
    }
    if(ifd<0) {
        Error("read_attach_fd: ignore ifd# %d", ifd);
    }

    *rifd = ifd;

    chk = bfd_seekobj(&ifd, bk, e, &npos, re);
    if(chk) {
        if(chk>0) {
            /*
             * MISS - if 'chk' is setted, it is not error.
             */
#if 0
            Trace("read_attach_fd: MISS, not found the object <%d>\n", chk);
#endif
            *rnpos = -1;
            ret = chk;
        }
        else {
            Error("read_attach_fd: fail bfd_seekobj() <%d>", chk);
            *rnpos = -1;
            ret = -1;
        }

        goto out;
    }

    *rnpos = npos;
    bpos = npos*Oblksize;

    chk = bfd_open(&bfd, bk, e, fszchk);
    if(chk) {
Trace("read_attach_fd: fail open bfd# %d, chk %d\n", bfd, chk);

        *rbfd = -1;
        ret = chk;

        goto out;
    }
    if(bfd<0) {
        Error("read_attach_fd: ignore bfd# %d", bfd);
    }
        

    *rbfd = bfd;

#ifdef NP
    bpos = (off_t) 0L;
#endif

    c = 0;
try_seek:
    c++;
    vpos = Pseek(fpos, bfd, bpos);
    if(vpos<0) {
Error("read_attach_fd: fail seek at %ld %lld in bfd# %d, vpos %d (%d)",
    (long)npos, (long long)bpos, bfd, vpos, errno);
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_seek;
        }
        *rnpos = -1;
        ret = 1;
    }
    else {
#if 0
Trace("read_attach_fd: done seek at pos %ld %lldB in bfd# %d, vpos %lldB\n",
    (long)npos, (long long)bpos, bfd, (long long)vpos);
#endif
        ret = 0;
    }


out:
#if 0
Trace("read_attach_fd: done *rifd# %d, *rbfd# %d, pos %ld %lldB, ret %d\n",
        *rifd, *rbfd, (long)*rnpos, (long long)bpos, ret);
#endif

    return ret;
}




int
detach_fd(int *ifd, int *bfd, bucket_t *bk, skcache_entry_t *e)
{
    int c1,c2;
    int sifd, sbfd;

    sifd = *ifd;
    sbfd = *bfd;

#ifdef TRACE_FD
    Trace("before detach_fd e %d:%d; ifd %d, bfd %d\n",
        e->h1, e->h2, *ifd, *bfd);
#endif

    if(*ifd>=0) {
        c1 = ifd_close(ifd, bk, e);
    }
    if(*bfd>=0) {
        c2 = bfd_close(bfd, bk, e);
    }

#if 0
    Trace("detach_fd: fd %d,%d, bk# %d, e %d:%d => c1,c2 %d,%d\n",
        sifd, sbfd, bk->bid, e->h1, e->h2, c1, c2);
#endif
#ifdef TRACE_FD
    Trace("after  detach_fd e %d:%d; ifd %d, bfd %d\n",
        e->h1, e->h2, *ifd, *bfd);
#endif
}



/*
 * return
 *      0   success
 *      >0  fail
 *
 */

int
skcache_entry_swapin(bucket_t *bk, skcache_entry_t *e)
{
    int ifd, bfd;
    int npos;
    skcache_entry_t tmp_entry;
    skcache_frag_t *cur, *pre, *dt;
    int nr;
    off_t nr_sum;
    int chk;
    int i;
    int ret;
    int nf;
    struct timeval begin, end;
    int c;
    off_t fsize;
    off_t tmpfsize;
    int erase;
    int sigpos;
    int lf;
    int er;
    int rest;

    int want;
    struct iovec iov[HARD_NFRAG_DREAD];
    skcache_frag_t *chunk[HARD_NFRAG_DREAD];
    off_t iovlen;
    int iovnum; 
    off_t tsum;

    int takes;
    int fsnum;


    gettimeofday(&begin, NULL);
    ret = 0;
    ifd = bfd = -87;

    FCHK();

#if 0
    Trace("swapin: b bk# %d, entry %p, oid# %d, d_recorded %d, m_recorded %p, count %d, len %d\n",
        bk->bid, e, e->oid, e->d_recorded, e->m_recorded, e->count, e->totallength);
#endif

    assert(sk_entity_check(bk, __FILE__, __LINE__));

    if(e->m_recorded!=NULL) {
        Error("swapin: not recorded in memory bk# %d, e %p", bk->bid, e);
#if 0
        ret = XSK_EIGNORE;
#endif
        ret = XSK_SWAP_EMPTY;
        goto out;
    }

    sigpos = DSIG_POS(e->h1, e->h2, e->h3);
    fsnum = bk->fs;

    MUTEX_LOCK(&stat_lock);
    swapin_count++;
    fs_swapin_count[fsnum]++;
    MUTEX_UNLOCK(&stat_lock);

    MUTEX_LOCK(&dir_mutex);
    lf = dir_lock[sigpos];
    MUTEX_UNLOCK(&dir_mutex);
    if(lf>0) {
/*
        Trace("swapin: conflict directory bk# %d, dir# %d,.\n", bk, sigpos);
*/
        MUTEX_LOCK(&stat_lock);
        dir_conflict++;
        MUTEX_UNLOCK(&stat_lock);

        ret = XSK_SWAP_ECONF;
    }
    if(ret) {
        goto out;
    }

    chk = read_attach_fd(&ifd, &bfd, &npos, bk, e, &tmp_entry);
    if(chk>0) {
#if 0
        Trace("swapin: not found object (chk=%d)\n", chk);
#endif
        ret = XSK_SWAP_MISS;
        detach_fd(&ifd, &bfd, bk, e);

        goto out;
    }
    if(chk<0 || ifd<0 || bfd<0 || ifd==bfd || npos<0) {
        Error("swapin: error chk %d, ifd# %d, bfd# %d, pos %ld",
            chk, ifd, bfd, (long)npos);
        ret = XSK_SWAP_EATTACH;
        goto out;
    }



    /***
     *** check contents
     ***/

    /*
     * this routing MUST check magic.
     * because disk maybe stores broken or older version cache file. 
     */
    if(tmp_entry.magic[0]!=MAGIC_0
        || tmp_entry.magic[1]!=MAGIC_1) {
#ifdef CHECK_MAGIC
        Error("swapin: ignore magic %02x %02x / %02x %02x '%s'",
            tmp_entry.magic[0], tmp_entry.magic[1],
            MAGIC_0, MAGIC_1, tmp_entry.name);
#endif
        ret = XSK_EMAGIC;
        goto fail_read_cntl;
    }

    tmp_entry.m_recorded = NULL;
#if 0
    Trace("swapin: read ifd# %d, npos %d, oid# %d, [%d:%d:%d:%d] %s\n",
        ifd, npos, 
        tmp_entry.oid, tmp_entry.h1, tmp_entry.h2,
        tmp_entry.h3, tmp_entry.h4, tmp_entry.name);
    Trace("  target is [%d] '%s'\n", e->h4, e->name);
#endif

    if(tmp_entry.oid<=0) {
#if 0
        Trace("swapin: reject negative oid in %p, oid# %d,\n",
            &tmp_entry, tmp_entry.oid);
#endif
        ret = XSK_EOID;
        goto fail_read_cntl;
    }

    if(tmp_entry.h4<0) {
#if 0
        Trace("swapin: reject negative h4 in %p, h4 %d,\n",
            &tmp_entry, tmp_entry.h4);
#endif
        ret = XSK_EH4;
        goto fail_read_cntl;
    }

    if(strcmp(tmp_entry.name,e->name)!=0) {
#if 0
        Trace("swapin: read oid# %d, hash %d,%d,%d,%d %s\n",
            tmp_entry.oid, tmp_entry.h1, tmp_entry.h2,
            tmp_entry.h3, tmp_entry.h4, tmp_entry.name);
        Trace("but target is '%s'\n", e->name);
#endif

#if 0
        Trace("    oid# %d, hash %d,%d,%d,%d\n",
            tmp_entry.oid, tmp_entry.h1, tmp_entry.h2,
            tmp_entry.h3, tmp_entry.h4);
        Trace("    saved %s\n    wont  %s\n", tmp_entry.name, e->name);
        {
            int yh1, yh2, yh3, yh4;
            calc_hash(tmp_entry.name,   &yh1, &yh2, &yh3, &yh4);
            calc_hash(e->name,          &yh1, &yh2, &yh3, &yh4);
        }
#endif

        ret = XSK_EHASH;
        erase = 1;
        goto fail_read_cntl;
    }

    if(tmp_entry.issue_time<0) {
#if 1
        Error("swapin: ignore issue_time");
#endif
        ret = XSK_EISSUE;
        goto fail_read_cntl;
    }

    if(tmp_entry.issue_time<min_date) {
#if 1
        Error("swapin: very old object");
#endif
        ret = XSK_EOLD;
        goto fail_read_cntl;
    }


    /***
     *** correct entry.
     ***/
    {
        int old_oid;
        void *old_owner;

        old_oid = e->oid;
        old_owner = (void*)e->owner;

        *e = tmp_entry;

        (void*)e->owner = old_owner;
        e->d_recorded = e->oid;
    }


    
skip_read_cntl:
    (void)1;


#if 0
    Trace("oid# %d, totallength %lld\n",
        e->oid, (long long)e->totallength);
#endif


    want = NBCEIL(e->totallength, Oblksize);

    for(i=0;i<HARD_NFRAG_DREAD;i++) {
        chunk[i] = NULL;
    }


    pre = NULL;

    nr_sum = (off_t)0L;
    nf = 0;                 /* the number of frag */
    do {
        rest = e->totallength - nr_sum;
#if 0
        Trace("oid# %d, rest %lld, totallength %lld, nr_sum %lld: dread %d\n",
            e->oid,
            (long long)rest, (long long)e->totallength, (long long)nr_sum,
            nfrag_dread);
#endif
        iovnum = 0;
        iovlen = 0L;
        for(i=0;i<nfrag_dread && iovlen < rest ;i++) {
            cur = skcache_pull_new_frag(bk);
            if(cur==NULL) {
                ret = XSK_NOFRAG;
                goto fail_read;
            }
            bk->own_frag++;

            chunk[i] = cur;
            iov[i].iov_base = cur->ent;
            if(iovlen + cur->size > rest) {
                iov[i].iov_len  = rest - iovlen;
                iovlen += iov[i].iov_len;
            }
            else {
                iov[i].iov_len  = cur->size;
                iovlen += cur->size;
            }
            iovnum++;
        }
        cur->next = NULL;

#if 0
    Trace("oid# %d, iov num %d len %lld\n",
        e->oid, iovnum, (long long)iovlen);
#endif

        c = 0;
try_read:
        nr = Preadv(sigpos, bfd, iov, iovnum);

#if 0
        Trace("<<< %3d %p: read %5d/%ld %7ld? oid# %d, %lld\n",
            nf, cur, nr, (long)iovlen,
            (nr>0) ? (long)nr_sum+nr : (long)nr_sum,
            e->oid, (long long)e->totallength);
#endif


        if(nr<0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_read;
            }

            Error("swapin: data, fail read() return %d %d (%d)",
                nr, c, errno);

#if 0
            cur->len = 0;
            skcache_push_idle_frag(bk, cur);
            bk->own_frag--;
            cur = NULL;
#endif

            ret = XSK_SWAP_EREAD;
            goto fail_read;
        }
#if 0
        if(nr==0) {
#if 0
            Trace("  oid# %d, reach end\n", e->oid);
#endif
            cur->len = 0;

            /*
             * When only read first frag, keep the frag.
             * otherwise, remove these frags to save memory.
             */
            if(nf>0) {
                skcache_push_idle_frag(bk, cur);
                bk->own_frag--;
                cur = NULL;
            }
            
            break;
        }
#endif

        if(nr!=iovlen) {
            Error("swapin: oid# %d, dir# %d, read act/spc %lld/%lld",
                e->oid, sigpos, (long long)nr, (long long)iovlen);
        }

        /* append read chunk to past read */
        tsum = 0;
        for(i=0;i<iovnum;i++) {
            cur = chunk[i];
            cur->len = iov[i].iov_len;

            if(pre==NULL) {
                if(e->m_recorded) {
Error("oid# %d, i over write %p upon %p", e->oid, cur, e->m_recorded);
                }
                e->m_recorded = cur;
            }
            else {
                pre->next = cur;
            }
            pre = cur;
            cur->next = NULL;

            nr_sum += iov[i].iov_len;
            tsum += iov[i].iov_len;


            if(tsum>=nr) {
                break;
            }
        }
        if(nr==0) {
            break;
        }

        nf++;
    } while( nr_sum < e->totallength );
 

#if 0
    Trace("  swapin 1 read %lld/%lld %lld oid# %d, '%s'\n",
        (long long)nr_sum, (long long)e->totallength,
        (long long)tmp_entry.totallength,
        e->oid, e->name);
#endif

    /*
     * fix length and empty frags in chunk
     */

#if 0
    Trace("  swapin 2 read %lld/%lld %lld oid# %d, '%s'\n",
        (long long)nr_sum, (long long)e->totallength,
        (long long)tmp_entry.totallength,
        e->oid, e->name);
#endif

    if(nr_sum != e->totallength) {
        Error("swapin: oid# %d, dir# %d, total read act/spc %lld/%lld '%s'",
            e->oid, sigpos,
            (long long)nr_sum, (long long)e->totallength,
            e->name);
    }

    MUTEX_LOCK(&stat_lock);
    swapin_success++;
    MUTEX_UNLOCK(&stat_lock);

    goto out;

/*
 * recoverely
 */

fail_read_cntl:
#if 0
    Trace("swapin: fail read contorl block reason %d, erase %d, size %ld\n",
        ret, erase, (long)fsize);
#endif
    goto fail;

fail_read:
#if 1
    Trace("swapin: fail to read data. trush just readings, reason %d\n",
        ret);
#endif

#if 0
    Trace("swapin: clear read %d frags, %lld bytes, reason %d\n", 
        nf, (long long)nr_sum, ret);
#endif

    goto fail;

fail:
    detach_fd(&ifd, &bfd, bk, e);
   
    if(ret<0) {
        sim_stat_add_value(&swapin_fail_reason, -ret);
    }
    else {
        sim_stat_add_value(&swapin_fail_reason, ret);
    }

    goto out;


out:
    detach_fd(&ifd, &bfd, bk, e);

#if 0
    Trace("  frag use/all = %d/%d\n", bk->own_frag, bk->frag_all);
#endif

    gettimeofday(&end, NULL);
    takes = KOTETU_SUB_TIME(end, begin);
    sim_stat_add_value(&swapin_time, takes);
    sim_hist_add_value(&swapin_hist, takes);
    sim_hist_add_value(&fs_swapin_hist[fsnum], takes);

#if 0
    Trace("swapin: a bk# %d, entry %p, oid# %d, d_recorded %d, m_recorded %p, count %d, len %d, return %d\n",
        bk->bid, e, e->oid, e->d_recorded, e->m_recorded, e->count, e->totallength, ret);
#endif

    FCHK();

    return ret;
}









/*
 *
 * return
 *      0   success
 *      !0  fail
 *              >0  ignore
 *              <0  temporary, redo again
 */

int
skcache_entry_swapout(bucket_t *bk, skcache_entry_t *e)
{
    int ret;
    int ifd, bfd;
    int npos;
    skcache_frag_t *cur, *pre;
    int nw;
    off_t nw_sum;
    int chk;
    int i, ii;
    struct timeval begin, end;
    int sig_pos;
    int c;
    int takes;

    int fsnum;

    struct iovec iov[HARD_NFRAG_DWRITE];
    int     numf;
    off_t   lenf;

    off_t last_size;
    int lf;

    int want;


    gettimeofday(&begin, NULL);

    ifd = bfd = -1;

    ret = 0;
    FCHK();

    if(bk==NULL) {
        Error("nil bucket pointer %x", bk);
        ret = XSK_EBUCKET;
        goto fail_start;
    }
    if(e==NULL) {
        Error("nil entry pointer %x", e);
        ret = XSK_EMPTY;
        goto fail_start;
    }

    sig_pos = DSIG_POS(e->h1, e->h2, e->h3);
    fsnum = bk->fs;

#if 0
    Trace("swapout: b bk# %d, entry %p, oid# %d, d_recorded %d, m_recorded %p, count %d, len %ld\n",
        bk->bid, e, e->oid, e->d_recorded, e->m_recorded, e->count, (long)e->totallength);
#endif

    assert(sk_entity_check(bk, __FILE__, __LINE__));

    if(e->m_recorded==NULL) {
      Trace("swapout: not recorded in memory bk# %d, e %p, oid# %d, length %ld\n",
            bk->bid, e, e->oid, (long)e->totallength);
        ret = 0;
        goto out;
    }
    if(e->issue_time==0 || e->oid==0) {
#if 0
        Trace("swapout: temporary entry bk# %d, e %p, oid# %d, length %ld\n",
            bk->bid, e, e->oid, (long)e->totallength);
#endif
        ret = 0;
        goto out;
    }

#ifndef SAVECOUNT
    if(e->d_recorded!=0)
    {
#if 0
    Trace("+# oid# %d, d_recorded# %d, count %3d,\n\thot %d, name %s\n",
        e->oid, e->d_recorded, e->count, e->hot, e->name);
#endif
        ret = 0;
        goto out;
    }
#endif

    if(e->totallength > Oblksize * Onblk_obj)
    {
#if 0
        Trace("swapout: oid# %d, file size %ld %ldK is larger than %ld <BLKx%d>\n",
            e->oid,
            (long)e->totallength+ESIZE,
            dK((long)e->totallength+ESIZE),
            (long)Oblksize*Onblk_obj, Onblk_obj);
#endif
        MUTEX_LOCK(&stat_lock);
        count_too_big_object++;
        MUTEX_UNLOCK(&stat_lock);

        ret = 0;
        goto out;
    }

    want = NBCEIL(e->totallength, Oblksize);

    MUTEX_LOCK(&stat_lock);
    swapout_count++;
    fs_swapout_count[fsnum]++;
    MUTEX_UNLOCK(&stat_lock);

    MUTEX_LOCK(&dir_mutex);
    lf = dir_lock[sig_pos];
    MUTEX_UNLOCK(&dir_mutex);
    if(lf>0) {
/*
        Trace("swapout: conflict directory bk# %d, dir# %d,.\n",
            bk->bid, sig_pos);
*/
        MUTEX_LOCK(&stat_lock);
        dir_conflict++;
        MUTEX_UNLOCK(&stat_lock);

        ret = XSK_SWAP_ECONF;
        goto fail_start;
    }

    chk = write_attach_fd(&ifd, &bfd, &npos, bk, e);
    if(chk || ifd<0 || bfd<0 || ifd==bfd || npos<0) {
/*
*/
#if 1
        Trace(
         "swapout: not found object (%d) or error ifd# %d, bfd# %d, pos %ld\n",
            chk, ifd, bfd, (long)npos);

        Trace("  dir# %d, dir_sz %lld %.2f%% hwater/max=%lld/%lld\n",
            sig_pos, (long long)dir_sz[sig_pos],
            ((double)dir_sz[sig_pos]*100.0/(double)dir_size_max),
            (long long)dir_size_hwater, (long long)dir_size_max
            );
#endif

        ret = XSK_SWAP_NOTFOUND;
        goto fail_open;
    }

#if 0
    Trace("swapout: oid# %d, attach to file ifd# %d, bfd# %d, \n",
        e->oid, ifd, bfd);
#endif

    /***
     *** write control
     ***/

#if 0
    if(want>1) {
        Trace("  *** want %d blocks over %d..%d\n", want, npos, npos+want);
    }
#endif
    for(ii=npos;ii<npos+want;ii++) {
        i = ii%Onblk_file;

        c = 0;
try_write_cntl1:
        c++;
        nw = Iwrite(sig_pos, ifd, e, ESIZE);
        if(nw<0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_write_cntl1;
            }

            if(errno==ENOSPC) {
                Error("disk full!! at write control in swapout 1");
            }
            else
            if(errno==ENOMEM) {
                Error("kernel said no memory. <%d>", c);

                msleep(1);

                if(c<DISK_RETRY) {
                    goto try_write_cntl1;
                }

            }
            else {
                Error("swo: fail Iwrite() index return %d %d (%d)", nw, c, errno);
            }
            ret = XSK_SWAP_EWRITE;
            goto fail_write_cntl;
        }

    }


    if(e->d_recorded!=0) 
    {
#if 0
    Trace("+# oid# %d, d_recorded# %d, count %3d,\n\thot %d, name %s\n",
        e->oid, e->d_recorded, e->count, e->hot, e->name);
#endif
        goto skip_body_swapout;
    }


    /***
     *** write body
     ***/

    cur = e->m_recorded;
    if(cur==NULL) {
        goto skip_body_swapout;
    }

    nw_sum = 0L;
    i = 0;
    while(cur) {

        /* write frag chunk by writev() */
        numf = 0;
        lenf = (off_t)0L;
        while(cur && numf<nfrag_dwrite) {
#if 0
            Trace(">>> %3d %3d %p: stored %5d/%d %7ld\n",
                i, numf, cur->ent, cur->len, cur->size, (long)nw_sum+cur->len);
#endif
            if(cur->len>0) {
                iov[numf].iov_base  = cur->ent;
                iov[numf].iov_len   = cur->len;
                numf++;

                lenf += (off_t) cur->len;
            }
            cur = cur->next;
        }
#if 1
        for(i=numf;i<HARD_NFRAG_DWRITE;i++) {
            iov[i].iov_base = NULL;
            iov[i].iov_len  = -1;
        }
#endif

        c = 0;
try_write:
        c++;
#if 0
        nw = writev(bfd, iov, numf);
#endif
        nw = Pwritev(sig_pos, bfd, iov, numf);
#if 0
        Trace(">>> %3d write %ld\n",
            i, nw);
#endif

        if(nw<0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_write;
            }
            Error("swo: fail Pwritev() data return %d (%d)", nw, errno);
            if(errno==ENOSPC) {
                Error("disk full!! when swapout data");
            }

            ret = 7;
            goto fail_write;
        }
        else
        if(nw==0) {
            break;
        }
        if(nw>0) {
#if 1
            if(nw!=lenf) {
                Error("swapout: oid# %d, write act/spc %lld/%lld bytes", 
                    e->oid, (long long)nw, (long long)lenf);
            }
#endif
            nw_sum += nw;
        }

        i++;
    }
#if 1
    if(nw_sum!=e->totallength) {
        Trace("swapout: oid# %d, total act/spc %lld/ %lld bytes\n",
            e->oid, (long long)nw_sum, (long long)e->totallength);
    }
#endif

    /* XXX */
#if 0
    ftruncate(bfd, (off_t) nw_sum + ESIZE );
#endif

skip_body_swapout:
    detach_fd(&ifd, &bfd, bk, e);
    skcache_entry_purge(bk, e);
    
    e->m_recorded = NULL;
    e->d_recorded = e->oid;
#if 1
    e->swapout_candidate = 0;
#endif

#if 0
    Trace("swapout: oid# %d, write %ld, %ld+%ld/%ld\n",
        e->oid, ESIZE + (long)nw_sum,
        ESIZE, (long)nw_sum, (long)e->totallength);
#endif

#if 0
    Trace("swapout: p %p, oid# %d, path '%s', size %ld\n", 
        e, e->oid, name, (long)nw_sum);
#endif

    last_size = 0;

    MUTEX_LOCK(&stat_lock);
    swapout_success++;

    if(last_size<=0) {
        swap_count += 1LL;
        swap_size +=
            (long long)(nw_sum + (off_t)ESIZE);
    }
    else {
        swap_size +=
            (long long)(nw_sum + (off_t)ESIZE - last_size);
    }
#if 0
    Trace("swapout: oid# %d, total swap count %lld size %lld\n",
        e->oid, swap_count, swap_size);
#endif
    MUTEX_UNLOCK(&stat_lock);

#if 0 
    MUTEX_LOCK(&dir_mutex);
#endif

    DSIG_SET(sig_pos, e->h4);

    for(ii=npos;ii<npos+want;ii++) {
        i = ii%Onblk_file;
        DBIT_SET(sig_pos, i);
        DMAP_SET(sig_pos, i, OMAP_FUNC(e));
    }

    if(dir_count[sig_pos]<0) {
        /* nothing */
    }
    else {
        dir_sz[sig_pos] +=
            (off_t)(nw_sum + (off_t)ESIZE - last_size);
        dir_count[sig_pos]++;
    }

#if 0
    MUTEX_UNLOCK(&dir_mutex);
#endif

#if 0
    dsig_show(&dir_sig[sig_pos*DSIG_ELEM_WIDTH]);
#endif
#if 0
    dbit_show(&dir_bit[sig_pos*DBIT_ELEM_WIDTH]);
#endif

    goto out;

fail_write_cntl:
fail_write:


    if(ret<0) {
        sim_stat_add_value(&swapout_fail_reason, -ret);
    }
    else {
        sim_stat_add_value(&swapout_fail_reason, ret);
    }

    MUTEX_LOCK(&stat_lock);
    if(last_size<0) {
    }
    else {
        swap_size -= last_size;
    }
    MUTEX_UNLOCK(&stat_lock);

    goto out;

fail_open:
fail_start:

    if(ret<0) {
        sim_stat_add_value(&swapout_fail_reason, -ret);
    }
    else {
        sim_stat_add_value(&swapout_fail_reason, ret);
    }

    goto out;

out:
    detach_fd(&ifd, &bfd, bk, e);
#if 0
    Trace("  frag use/all = %d/%d\n", bk->own_frag, bk->frag_all);
#endif
    gettimeofday(&end, NULL);
    takes = KOTETU_SUB_TIME(end, begin);
    sim_stat_add_value(&swapout_time, takes);
    sim_hist_add_value(&swapout_hist, takes);
    sim_hist_add_value(&fs_swapout_hist[fsnum], takes);

    assert(sk_entity_check(bk, __FILE__, __LINE__));
#if 0
    Trace("swapout: a bk# %d, entry %p, oid# %d, d_recorded %d, m_recorded %p, count %d, len %ld, return %d\n",
        bk->bid, e, e->oid, e->d_recorded, e->m_recorded, e->count, (long)e->totallength, ret);
#endif

    FCHK();

    return ret;
}





#ifdef NO_VS
#ifdef VS
#undef VS
#endif
#endif



extern pthread_mutex_t stat_lock;
extern volatile int compaction_count;

typedef struct {
    int no;
    int type;   /* 0 single, 1 packed */
    int pos;    /* position */
    int act;
    time_t date;
    off_t sz;
} tup;

int
comp_tup(const void* xp, const void* yp)
{
    tup *x = (tup*)xp;
    tup *y = (tup*)yp;

    if(x->act == y->act) {
        if(x->date == y->date) {
            return x->sz - y->sz;           /* sz smaller */
        }
        else {
            return y->date - x->date;       /* date grater */
        }
    }
    else {
        return x->act - y->act;             /* act smaller */
    }
}

int
comp_tup_order(const void* xp, const void* yp)
{
    tup *x = (tup*)xp;
    tup *y = (tup*)yp;

    if(x->type == y->type) {
        return x->pos - y->pos;             /* y-x smaller */
    }
    else {
        return y->type- x->type;            /* y-x greater */
    }
}

int
comp_tup_time(void*xp,void*yp)
{
    tup *x = (tup*)xp;
    tup *y = (tup*)yp;

    return x->date - y->date;
}

int
show_tup(char *msg, int num, tup *base)
{
    int i;
    tup *cur;
    off_t lsum, osum;

    Trace("show_tup: %s\n", msg);

    lsum = osum = (off_t) 0L;
    cur = base;
    for(i=0;i<num;i++) {
        if(cur->act==ACTION_NOP || cur->act==ACTION_SKIP) {
			lsum += cur->sz;
			Trace(" %12d: %c %4d : %d %10d %12lld %12lld\n",
				cur->no, cur->type ? 'P' : 's', cur->pos,
				cur->act, cur->date, (long long)cur->sz, (long long)lsum);
        }
        else {
			osum += cur->sz;
			Trace(" %12d: %c %4d : %d %10d %12lld\n",
				cur->no, cur->type ? 'P' : 's', cur->pos,
				cur->act, cur->date, (long long)cur->sz);
        }
        cur++;
    }
    Trace("                                                  %12lld\n",
		(long long)lsum);
}


int 
decision_disk_action(skcache_entry_t *en, time_t reftime, int *xaction)
{
    int xh1, xh2, xh3, xh4;
    int igh;

    *xaction = ACTION_SKIP;

#if 0
Trace("check_target: en %p, MAGIC %x %x, oid# %d, *xaction %d\n",
    en, (unsigned char)en->magic[0], (unsigned char)en->magic[1],
    en->oid, *xaction);
#endif

    if(en->magic[0]!=MAGIC_0 || en->magic[1]!=MAGIC_1) {
        *xaction = ACTION_PURGE;
#ifdef CHECK_MAGIC
Error("check_target: oid# %d, wrong magic %02x %02x / %02x %02x",
        en->oid, 
        en->magic[0], en->magic[1], MAGIC_0, MAGIC_1);
#endif
        en->issue_time = -1;
        goto out;
    }

    if(en->oid == 0) {
#if 0
Trace("oid# %d, *xaction %d -> 0\n", en->oid, *xaction);
#endif
        *xaction = ACTION_PURGE;
        goto out;
    }

    calc_hash(en->name, &xh1, &xh2, &xh3, &xh4);

    if(en->h1!=xh1 || en->h2!=xh2 || en->h3!=xh3 || en->h4!=xh4) {
        *xaction = ACTION_PURGE;
        igh++;
#if 0
Trace("\
check_target: ignore hash, file '%s'\n\
    stored %03d:%03d:%03d:%03d %s\n\
    solved %03d:%03d:%03d:%03d same\n",
    name,
    en->h1, en->h2, en->h3, en->h4, en->name,
    xh1, xh2, xh3, xh4);
#endif
#if 0
Trace("check_target: ignore hash, stored %03d:%03d:%03d:%03d, we guess %03d:%03d:%03d:%03d\n",
        en->h1, en->h2, en->h3, en->h4, xh1, xh2, xh3, xh4);
#endif
        en->issue_time = -1;
        goto out;
    }

/*
    idiff = reftime - en->issue_time;
    ldiff = reftime - en->lmd;
    ediff = reftime - en->expire;
*/


    if(en->issue_time<=0) {
        *xaction = ACTION_PURGE;
    }
    else {
        /* black */
        if(en->issue_time<min_date) {
            *xaction = ACTION_PURGE;
        }
        else
        if(en->issue_time>reftime) {
            *xaction = ACTION_PURGE;
        }
        else
        if(en->expire>0 && reftime >= en->expire ) {
            *xaction = ACTION_PURGE;
        }

        /* gray */
    }

out:
    return *xaction;
}


int check_target(skcache_entry_t *en, time_t reftime, int *xaction)
{
    return decision_disk_action(en, reftime, xaction);
}


/*
 * convert num-num -> num
 */
int name2num(char *nm)
{
    int r1, r2, r3, r4;
    int ret;
    register char *p, *q;
    register int i;
    char tmp[L_PATH];
    
    p = nm;

    q = tmp;
    i = 0;
    while(*p && (*p>='0' && *p<='9') && i < L_PATH) {
        *q++ = *p++;
        i++;
    }
    p++;
    *q = '\0';
    r3 = atoi(tmp);
    
    q = tmp;
    i = 0;
    while(*p && (*p>='0' && *p<='9') && i < L_PATH) {
        *q++ = *p++;
        i++;
    }
    p++;
    *q = '\0';
    r4 = atoi(tmp);

    ret = r3*radix_h4 + r4;

#if 0
    Trace("name2num '%s' -> %d\n", nm, ret);
#endif

    return ret;
}


#include "chkdir.c"


volatile int bsweeper_seq   = 0;
volatile int bsweeper_count = 0;
volatile int bsweeper_phase = -1;
volatile int bsweeper_pos   = -1;
volatile int bsweeper_full_mode = 0;

/* how many tics are required to work sweeper */
#ifndef SWEEP_CYCLE
#define SWEEP_CYCLE (1)
#endif
#ifndef SWEEP_BOOT_CYCLE
#define SWEEP_BOOT_CYCLE (1)
#endif

/* how many direcotry are sweeped per chance */
#ifndef SWEEP_NTRY
#define SWEEP_NTRY  (1)
#endif


volatile int bsweep_boot_cycle = SWEEP_BOOT_CYCLE;
volatile int bsweep_cycle = SWEEP_CYCLE;



/* for two level */
int
skcache_dir_sweeper(void)
{
    int f;
    int d1, d2, d3;
    int v;
    int chk;
    int p;
    int nd1;
    int pos;

    MUTEX_LOCK(&stat_lock);
    if(bsweeper_phase<0) {
        bsweeper_phase = 0;
        bsweeper_pos = rand() % dir_num;
    }
    p = bsweeper_phase;
    f = bsweeper_seq;
    bsweeper_seq =  (bsweeper_seq+1) % 1000000;

#if 0
    if(bsweeper_count%dir_num==0) {
        Trace("first dir\n");
    }
#endif
    bsweeper_count++;

    v = bsweeper_pos;

    if(bsweeper_pos>=dir_num-1) {
        bsweeper_phase++;
        bsweeper_pos = 0;
    }
    else {
        bsweeper_pos++;
    }
    MUTEX_UNLOCK(&stat_lock);

    if(f<0) {
        f = v;
    }

    d1  = v%radix_h1;
    d2  = v/radix_h1;
    d3  = 0;

    pos = DSIG_POS(d1, d2, d3);

#if 0
    Trace("  sweeper th %08x %6d %6d -> dir# %d, %03d %03d %03d\n",
        pthread_self(), p, v, pos, d1, d2, d3);
#endif
    MUTEX_LOCK(&dir_mutex);
    if(dir_lock[pos]>0) {
        pos = -1;
    }
    else {
        dir_lock[pos] = pthread_self();
    }
    MUTEX_UNLOCK(&dir_mutex);

    if(pos>=0) {

        chk = skcache_check_dir(d1, d2, d3, 1, 1, 0);

        MUTEX_LOCK(&dir_mutex);
        dir_lock[pos] = 0;
        MUTEX_UNLOCK(&dir_mutex);
    }

done:

    return 0;
}


static
void
skcache_preload_files()
{
    int i, j;
    skcache_entry_t dummy_ent;
    int dummy_ifd, dummy_bfd;
    int chk;

    Trace("skcache_preload_files: preload %d files (%d x %d x 2)\n",
        radix_h1 * radix_h2 * 2, radix_h1, radix_h2);
    for(i=0;i<radix_h1;i++) {
        for(j=0;j<radix_h2;j++) {
            dummy_ent.h1 = i;
            dummy_ent.h2 = j;

            chk = ifd_open(&dummy_ifd, NULL, &dummy_ent, fszchk);
            chk = bfd_open(&dummy_bfd, NULL, &dummy_ent, fszchk);
#if 0
            Trace(" (i,j) %4d %4d -> (ifd, bfd) %5d %5d\n",
                i, j, dummy_ifd, dummy_bfd);
#endif
        }
    }

}



void
skcache_disk_setup()
{
    int i, j;
    long long tmp;
    int fdpoolsize = radix_h1*radix_h2;

    if(cache_index_prefix==NULL || *cache_index_prefix=='\0') {
        cache_index_prefix = cache_dir_prefix;
    }

    if(radix_disk<2) {
        for(i=0;i<nbucket;i++) {
            bk2fs[i] = 0;
        }
    }
    else {
        i = chk_bk2fs_map(0);
        Trace("chk fsmap ret %d\n", i);
        if(i) {
            Error("fsmap is wrong");
            fflush(stdout);
            fflush(stderr);
            /*
            show_bk2fs_map();
            */
            show_fs2bk_map();
            fflush(stdout);
            fflush(stderr);

            equlize_fsmap();

            i = chk_bk2fs_map(0);
            Trace("chk fsmap ret %d\n", i);
        }
    }

    fflush(stdout);
    fflush(stderr);
    show_fs2bk_map();
#if 0
    show_bk2fs_map();
#endif
    fflush(stderr);
    fflush(stdout);

    Oblksize    = require_blksize;
    Onblk_file  = require_nblk_file;
    Onblk_seek  = require_nblk_seek;
    Onblk_obj   = require_nblk_obj;


    bfd_map = (int*)MALLOC(sizeof(int)*fdpoolsize);
    ifd_map = (int*)MALLOC(sizeof(int)*fdpoolsize);
    mlist_add_elem("index fds",(void*)ifd_map, sizeof(int)*fdpoolsize);
    mlist_add_elem("body  fds",(void*)bfd_map, sizeof(int)*fdpoolsize);

    ifd_mod = (int*)MALLOC(sizeof(int)*fdpoolsize);
    bfd_mod = (int*)MALLOC(sizeof(int)*fdpoolsize);
    mlist_add_elem("index mod",(void*)ifd_mod, sizeof(int)*fdpoolsize);
    mlist_add_elem("body  mod",(void*)bfd_mod, sizeof(int)*fdpoolsize);

    ifd_ref = (int*)MALLOC(sizeof(int)*fdpoolsize);
    bfd_ref = (int*)MALLOC(sizeof(int)*fdpoolsize);
    mlist_add_elem("index ref",(void*)ifd_ref, sizeof(int)*fdpoolsize);
    mlist_add_elem("body  ref",(void*)bfd_ref, sizeof(int)*fdpoolsize);

    ifd_pos = (off_t*)MALLOC(sizeof(off_t)*fdpoolsize);
    bfd_pos = (off_t*)MALLOC(sizeof(off_t)*fdpoolsize);
    mlist_add_elem("index pos",(void*)ifd_pos, sizeof(off_t)*fdpoolsize);
    mlist_add_elem("body  pos",(void*)bfd_pos, sizeof(off_t)*fdpoolsize);


    for(i=0;i<fdpoolsize;i++) {
        ifd_map[i] = bfd_map[i] =  -1;
        ifd_ref[i] = bfd_ref[i] =   0;
        ifd_mod[i] = bfd_mod[i] =   0;
        ifd_pos[i] = bfd_pos[i] = (off_t)-1L;
    }

    dir_size_max    = Oblksize*Onblk_file;
    dir_size_hwater = dir_size_max*19/20;
    dir_size_lwater = dir_size_max*16/20;

    /* change default values */
    fd_comp_width = 128;
#if 0
    nfd_cached_hwater  = 12800;
#endif
    nfd_cached_hwater  = 5120;
    nfd_cached_hhwater = nfd_cached_hwater*3/2;
    nfd_cached_lwater  = nfd_cached_hwater - fd_comp_width;
    ifd_keeptime = 4;
    bfd_keeptime = 2;




    dir_bit = (u_int32_t*)MALLOC(sizeof(u_int32_t)*dir_num*DBIT_ELEM_WIDTH);
    mlist_add_elem("dir_bit", (void*)dir_bit,
        sizeof(u_int32_t)*dir_num*DBIT_ELEM_WIDTH);

    dir_map = (int*)MALLOC(sizeof(int)*fdpoolsize*Onblk_file);
    mlist_add_elem("dir_map", dir_map, sizeof(int)*fdpoolsize*Onblk_file);


    for(i=0;i<fdpoolsize;i++) {
        for(j=0;j<Onblk_file;j++) {
            DMAP_SET(i, j, -1);
            DBIT_UNSET(i, j);
        }
    }

    disk_sweep_interval = require_disk_sweep_interval;




    sim_stat_U_alloc_new(&swapin_time,  "swapin time [msec]",
        10, K(1),   swapin_time_count);
    sim_stat_U_alloc_new(&swapout_time,   "swapout time [msec]",
        10, K(1),   swapout_time_count);

    sim_stat_U_alloc_new(&checkdir_time,    "checkdir time [msec]",
        10, K(2),   checkdir_time_count);
    sim_stat_U_alloc_new(&checkdir_time_short, "checkdir time (short) [msec]",
        10, K(2),   checkdir_time_short_stat_count);

    sim_stat_alloc_new(&swapout_fail_reason,    "swapout fail reason",
        K(1),   swapout_fail_reason_count);
    sim_stat_alloc_new(&swapin_fail_reason,   "swapin fail reason",
        K(1),   swapin_fail_reason_count);

    sim_hist_alloc_new(&swapin_hist,    "swapin histroy [msec]",
        K(1), swapin_hist_area);
    sim_hist_alloc_new(&swapout_hist,   "swapout histroy [msec]",
        K(1), swapout_hist_area);

    if(radix_disk<=1) {
        j = 1;
    }
    else {
        j = radix_disk;
    }
    for(i=0;i<j;i++) {
        fs_swapin_count[i] = 0;
        fs_swapout_count[i] = 0;
        sim_hist_alloc_new(&fs_swapin_hist[i],  "fs swapin histroy [msec]",
            FS_HIST_LEN, fs_swapin_hist_area[i]);
        sim_hist_alloc_new(&fs_swapout_hist[i], "fs swapout histroy [msec]",
            FS_HIST_LEN, fs_swapout_hist_area[i]);
    }



    /*
     * print parameters
     */
    Trace("\
WWW Object Caching Area upon Disk(Filesystem)s:\n\
  index '%s'\n\
  body  '%s'\n\
\n",
        cache_index_prefix, cache_index_prefix);

#ifdef PP
    Trace("PP is defined\n");
#endif
#ifdef NP
    Trace("NP is defined\n");
#endif

    Trace("\
Object Packing:\n\
  The Size of Block <Oblksize>                      %7d\n\
  The Number of Blocks per Object File <Onblk_file> %7d\n\
  The Number of Blocks per Seek Width <Onblk_seek>  %7d\n\
  The Number of Blocks per Object <Onblk_obj>       %7d range [0..%d)\n\
\n",
        Oblksize, Onblk_file, Onblk_seek,
        Onblk_obj, Oblksize*Onblk_obj);
    Trace("\
Disk:\n\
    %8d dirs   ; %d x %d\n\
    %8d files  ; %d file pairs (index and body)\n\
    %8d blocks ; %d x %d x %d\n\
              %8s %5s %10s %7s %14s %7s\n\
        INDEX %8ld %5ld %10lld %7.2f %14lld %7.2f\n\
        BODY  %8ld %5ld %10lld %7.2f %14lld %7.2f\n\
    Allocated %lld Bytes %.2f GB\n\
\n\
    water mark of dir\n\
        high %lld %.2fMB\n\
        low  %lld %.2fMB\n\
\n",
    dir_num, radix_h1, radix_h2,
    dir_num*2, dir_num,
    dir_num*Onblk_file, radix_h1, radix_h2, Onblk_file,
    "block",  "xN", "single", "MB", "all", "GB",
    ESIZE, Onblk_file, 
        (long long) ESIZE * Onblk_file, 
        fdM(ESIZE * Onblk_file), 
        (long long) dir_num * ESIZE * Onblk_file, 
        fdG((long long) dir_num * ESIZE * Onblk_file), 
    Oblksize, Onblk_file, 
        (long long) Oblksize * Onblk_file, 
        fdM( Oblksize * Onblk_file), 
        (long long) dir_num * Oblksize * Onblk_file, 
        fdG((long long) dir_num * Oblksize * Onblk_file),
    (long long)dir_num * (ESIZE+Oblksize) * Onblk_file,
        fdG((long long)dir_num * (ESIZE+Oblksize) * Onblk_file),
    (long long)dir_size_hwater,
        fdM((long long)dir_size_hwater),
    (long long)dir_size_lwater,
        fdM((long long)dir_size_lwater)
    );

    Trace("\n\
FD caching: %s\n\
    the number of files in fileset %6d\n\
    water mark hhigh %6d (%5.2f%% of fileset)\n\
               high  %6d (%5.2f%% of fileset)\n\
               low   %6d (%5.2f%% of fileset)\n\
    swepping width %d\n\
    keeptime index %d, body %d\n\
\n",
    (fdcaching_flag==1) ? "ON" : "OFF",
    2*fdpoolsize,
    nfd_cached_hhwater, 100.0*nfd_cached_hhwater/(2.0*fdpoolsize),
    nfd_cached_hwater,  100.0*nfd_cached_hwater/(2.0*fdpoolsize),
    nfd_cached_lwater,  100.0*nfd_cached_lwater/(2.0*fdpoolsize),
    fd_comp_width, ifd_keeptime, bfd_keeptime);

    Trace("\
Bitmapping: %s\n",
        bitmapping_flag==1 ? "ON" : "OFF");

    Trace("\
Bitmaps for directories:\n\
  SIZE_OF_UINT        %4d\n\
  DBIT_BIT_WIDTH      %4d\n\
  DBIT_VALUE_RANGE    %4d\n\
  DBIT_VALUE_RES      %4d\n\
  DBIT_ELEM_WIDTH     %4d\n\
  DBIT_BYTE_WIDTH     %4d\n\
\n",
        SIZE_OF_UINT,
        DBIT_BIT_WIDTH, DBIT_VALUE_RANGE, DBIT_VALUE_RES,
        DBIT_ELEM_WIDTH, DBIT_BYTE_WIDTH);

#if 0
    skcache_preload_files();
#endif


}

void
skcache_disk_init()
{
    TRACE_RECEIPT;

    clear_bk2fs_map();
}


